@@ -1,4 +1,5 @@
 #include "libretro.h"
+#include "libretro_core_options.h"
 
 #include "snes9x.h"
 #include "memmap.h"
@@ -23,6 +24,7 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <fcntl.h>
+#include "filter/snes_ntsc.h"
 
 #define RETRO_DEVICE_JOYPAD_MULTITAP ((1 << 8) | RETRO_DEVICE_JOYPAD)
 #define RETRO_DEVICE_LIGHTGUN_SUPER_SCOPE ((1 << 8) | RETRO_DEVICE_LIGHTGUN)
@@ -54,6 +56,8 @@ uint16 *screen_buffer = NULL;
 char g_rom_dir[1024];
 char g_basename[1024];
 
+bool g_geometry_update = false;
+
 int hires_blend = 0;
 bool randomize_memory = false;
 
@@ -67,6 +71,15 @@ static retro_audio_sample_batch_t audio_batch_cb = NULL;
 static retro_input_poll_t poll_cb = NULL;
 static retro_input_state_t input_state_cb = NULL;
 
+static snes_ntsc_t *snes_ntsc = NULL;
+static int blargg_filter = 0;
+static uint16 *ntsc_screen_buffer, *snes_ntsc_buffer;
+
+const int MAX_SNES_WIDTH_NTSC = ((SNES_NTSC_OUT_WIDTH(256) + 3) / 4) * 4;
+
+static bool show_lightgun_settings = true;
+static bool show_advanced_av_settings = true;
+
 static void extract_basename(char *buf, const char *path, size_t size)
 {
     const char *base = strrchr(path, '/');
@@ -133,6 +146,7 @@ enum overscan_mode {
 };
 enum aspect_mode {
     ASPECT_RATIO_4_3,
+    ASPECT_RATIO_4_3_SCALED,
     ASPECT_RATIO_1_1,
     ASPECT_RATIO_NTSC,
     ASPECT_RATIO_PAL,
@@ -143,6 +157,22 @@ static overscan_mode crop_overscan_mode = OVERSCAN_CROP_ON; // default to crop
 static aspect_mode aspect_ratio_mode = ASPECT_RATIO_4_3; // default to 4:3
 static bool rom_loaded = false;
 
+enum lightgun_mode
+{
+	SETTING_GUN_INPUT_LIGHTGUN,
+	SETTING_GUN_INPUT_POINTER
+};
+static lightgun_mode setting_gun_input = SETTING_GUN_INPUT_LIGHTGUN;
+
+// Touchscreen sensitivity vars
+static int pointer_pressed = 0;
+static const int POINTER_PRESSED_CYCLES = 4;
+static int pointer_cycles_after_released = 0;
+static int pointer_pressed_last_x = 0;
+static int pointer_pressed_last_y = 0;
+
+static bool setting_superscope_reverse_buttons = false;
+
 void retro_set_environment(retro_environment_t cb)
 {
     environ_cb = cb;
@@ -167,50 +197,7 @@ void retro_set_environment(retro_environment_t cb)
 
     cb(RETRO_ENVIRONMENT_SET_SUBSYSTEM_INFO,  (void*)subsystems);
 
-
-    struct retro_variable variables[] = {
-        // These variable names and possible values constitute an ABI with ZMZ (ZSNES Libretro player).
-        // Changing "Show layer 1" is fine, but don't change "layer_1"/etc or the possible values ("Yes|No").
-        // Adding more variables and rearranging them is safe.
-        { "snes9x_up_down_allowed", "Allow Opposing Directions; disabled|enabled" },
-        { "snes9x_overclock_superfx", "SuperFX Overclocking; 100%|150%|200%|250%|300%|350%|400%|450%|500%|50%|60%|70%|80%|90%" },
-        { "snes9x_overclock_cycles", "Reduce Slowdown (Hack, Unsafe); disabled|light|compatible|max" },
-        { "snes9x_reduce_sprite_flicker", "Reduce Flickering (Hack, Unsafe); disabled|enabled" },
-        { "snes9x_randomize_memory", "Randomize Memory (Unsafe); disabled|enabled" },
-        { "snes9x_hires_blend", "Hires Blending; disabled|merge|blur" },
-        { "snes9x_audio_interpolation", "Audio Interpolation; gaussian|cubic|sinc|none|linear" },
-        { "snes9x_layer_1", "Show layer 1; enabled|disabled" },
-        { "snes9x_layer_2", "Show layer 2; enabled|disabled" },
-        { "snes9x_layer_3", "Show layer 3; enabled|disabled" },
-        { "snes9x_layer_4", "Show layer 4; enabled|disabled" },
-        { "snes9x_layer_5", "Show sprite layer; enabled|disabled" },
-        { "snes9x_gfx_clip", "Enable graphic clip windows; enabled|disabled" },
-        { "snes9x_gfx_transp", "Enable transparency effects; enabled|disabled" },
-        { "snes9x_gfx_hires", "Enable hires mode; enabled|disabled" },
-        { "snes9x_sndchan_1", "Enable sound channel 1; enabled|disabled" },
-        { "snes9x_sndchan_2", "Enable sound channel 2; enabled|disabled" },
-        { "snes9x_sndchan_3", "Enable sound channel 3; enabled|disabled" },
-        { "snes9x_sndchan_4", "Enable sound channel 4; enabled|disabled" },
-        { "snes9x_sndchan_5", "Enable sound channel 5; enabled|disabled" },
-        { "snes9x_sndchan_6", "Enable sound channel 6; enabled|disabled" },
-        { "snes9x_sndchan_7", "Enable sound channel 7; enabled|disabled" },
-        { "snes9x_sndchan_8", "Enable sound channel 8; enabled|disabled" },
-        { "snes9x_overscan", "Crop overscan; enabled|disabled|auto" },
-        { "snes9x_aspect", "Preferred aspect ratio; 4:3|uncorrected|auto|ntsc|pal" },
-        { "snes9x_region", "Console region (Reload core); auto|ntsc|pal" },
-        { "snes9x_superscope_crosshair", "Super Scope crosshair; 2|3|4|5|6|7|8|9|10|11|12|13|14|15|16|0|1" },
-        { "snes9x_superscope_color", "Super Scope color; White|White (blend)|Red|Red (blend)|Orange|Orange (blend)|Yellow|Yellow (blend)|Green|Green (blend)|Cyan|Cyan (blend)|Sky|Sky (blend)|Blue|Blue (blend)|Violet|Violet (blend)|Pink|Pink (blend)|Purple|Purple (blend)|Black|Black (blend)|25% Grey|25% Grey (blend)|50% Grey|50% Grey (blend)|75% Grey|75% Grey (blend)" },
-        { "snes9x_justifier1_crosshair", "Justifier 1 crosshair; 4|5|6|7|8|9|10|11|12|13|14|15|16|0|1|2|3" },
-        { "snes9x_justifier1_color", "Justifier 1 color; Blue|Blue (blend)|Violet|Violet (blend)|Pink|Pink (blend)|Purple|Purple (blend)|Black|Black (blend)|25% Grey|25% Grey (blend)|50% Grey|50% Grey (blend)|75% Grey|75% Grey (blend)|White|White (blend)|Red|Red (blend)|Orange|Orange (blend)|Yellow|Yellow (blend)|Green|Green (blend)|Cyan|Cyan (blend)|Sky|Sky (blend)" },
-        { "snes9x_justifier2_crosshair", "Justifier 2 crosshair; 4|5|6|7|8|9|10|11|12|13|14|15|16|0|1|2|3" },
-        { "snes9x_justifier2_color", "Justifier 2 color; Pink|Pink (blend)|Purple|Purple (blend)|Black|Black (blend)|25% Grey|25% Grey (blend)|50% Grey|50% Grey (blend)|75% Grey|75% Grey (blend)|White|White (blend)|Red|Red (blend)|Orange|Orange (blend)|Yellow|Yellow (blend)|Green|Green (blend)|Cyan|Cyan (blend)|Sky|Sky (blend)|Blue|Blue (blend)|Violet|Violet (blend)" },
-        { "snes9x_rifle_crosshair", "M.A.C.S. rifle crosshair; 2|3|4|5|6|7|8|9|10|11|12|13|14|15|16|0|1" },
-        { "snes9x_rifle_color", "M.A.C.S. rifle color; White|White (blend)|Red|Red (blend)|Orange|Orange (blend)|Yellow|Yellow (blend)|Green|Green (blend)|Cyan|Cyan (blend)|Sky|Sky (blend)|Blue|Blue (blend)|Violet|Violet (blend)|Pink|Pink (blend)|Purple|Purple (blend)|Black|Black (blend)|25% Grey|25% Grey (blend)|50% Grey|50% Grey (blend)|75% Grey|75% Grey (blend)" },
-        { "snes9x_block_invalid_vram_access", "Block Invalid VRAM Access; enabled|disabled" },
-        { NULL, NULL },
-    };
-
-    environ_cb(RETRO_ENVIRONMENT_SET_VARIABLES, variables);
+    libretro_set_core_options(environ_cb);
 
     static const struct retro_controller_description port_1[] = {
         { "None", RETRO_DEVICE_NONE },
@@ -300,9 +287,10 @@ char *get_cursor_color(const char *name)
 		lcv += 2;
 	}
 
-	return "None";
+	return color_names[16]; // White
 }
 
+// always ensure this is only called in retro_run
 void update_geometry(void)
 {
     struct retro_system_av_info av_info;
@@ -310,11 +298,11 @@ void update_geometry(void)
     environ_cb(RETRO_ENVIRONMENT_SET_GEOMETRY, &av_info);
     g_screen_gun_width = av_info.geometry.base_width;
     g_screen_gun_height = av_info.geometry.base_height;
+    g_geometry_update = false;
 }
 
 static void update_variables(void)
 {
-    bool geometry_update = false;
     char key[256];
     struct retro_variable var;
 
@@ -460,7 +448,7 @@ static void update_variables(void)
         if (newval != crop_overscan_mode)
         {
             crop_overscan_mode = newval;
-            geometry_update = true;
+            g_geometry_update = true;
         }
     }
 
@@ -475,13 +463,15 @@ static void update_variables(void)
             newval = ASPECT_RATIO_PAL;
         else if (strcmp(var.value, "4:3") == 0)
             newval = ASPECT_RATIO_4_3;
+        else if (strcmp(var.value, "4:3 scaled") == 0)
+            newval = ASPECT_RATIO_4_3_SCALED;
         else if (strcmp(var.value, "uncorrected") == 0)
             newval = ASPECT_RATIO_1_1;
 
         if (newval != aspect_ratio_mode)
         {
             aspect_ratio_mode = newval;
-            geometry_update = true;
+            g_geometry_update = true;
         }
     }
 
@@ -506,6 +496,22 @@ static void update_variables(void)
         }
     }
 
+    var.key="snes9x_lightgun_mode";
+    if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE, &var))
+    {
+      if ( !strcmp(var.value, "Touchscreen") ) {
+         setting_gun_input = SETTING_GUN_INPUT_POINTER;
+      } else {
+         setting_gun_input = SETTING_GUN_INPUT_LIGHTGUN;
+      }
+    }
+
+    var.key="snes9x_superscope_reverse_buttons";
+    if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE, &var))
+    {
+        setting_superscope_reverse_buttons = strcmp(var.value, "enabled") == 0;
+    }
+
     var.key="snes9x_superscope_crosshair";
 
     if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE, &var))
@@ -602,8 +608,147 @@ static void update_variables(void)
     else
         Settings.BlockInvalidVRAMAccessMaster = true;
 
-    if (geometry_update)
-        update_geometry();
+    var.key = "snes9x_echo_buffer_hack";
+    var.value = NULL;
+
+    if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE, &var))
+        Settings.SeparateEchoBuffer = !strcmp(var.value, "disabled") ? false : true;
+    else
+        Settings.SeparateEchoBuffer = false;
+
+    var.key = "snes9x_blargg";
+    var.value = NULL;
+
+    if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE, &var))
+    {
+        if (strcmp(var.value, "disabled") == 0)
+            blargg_filter = 0;
+        else
+        {
+            int old_filter = blargg_filter;
+
+            if(!snes_ntsc) snes_ntsc = new snes_ntsc_t;
+            snes_ntsc_setup_t setup = snes_ntsc_composite;
+
+            if (strcmp(var.value, "monochrome") == 0)
+            {
+                blargg_filter = 1;
+
+                setup = snes_ntsc_monochrome;
+            }
+            else if (strcmp(var.value, "rf") == 0)
+            {
+                blargg_filter = 2;
+
+                setup = snes_ntsc_composite;
+                setup.merge_fields = 0;
+            }
+            else if (strcmp(var.value, "composite") == 0)
+            {
+                blargg_filter = 3;
+
+                setup = snes_ntsc_composite;
+            }
+            else if (strcmp(var.value, "s-video") == 0)
+            {
+                blargg_filter = 4;
+
+                setup = snes_ntsc_svideo;
+            }
+            else if (strcmp(var.value, "rgb") == 0)
+            {
+                blargg_filter = 5;
+
+                setup = snes_ntsc_rgb;
+            }
+
+            if (old_filter != blargg_filter)
+                snes_ntsc_init( snes_ntsc, &setup );
+        }
+    }
+
+    /* Show/hide core options */
+
+    var.key = "snes9x_show_lightgun_settings";
+    var.value = NULL;
+
+    if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE, &var) && var.value)
+    {
+        bool show_lightgun_settings_prev = show_lightgun_settings;
+
+        show_lightgun_settings = true;
+        if (strcmp(var.value, "disabled") == 0)
+            show_lightgun_settings = false;
+
+        if (show_lightgun_settings != show_lightgun_settings_prev)
+        {
+            size_t i;
+            struct retro_core_option_display option_display;
+            char lightgun_keys[10][64] = {
+                "snes9x_lightgun_mode",
+                "snes9x_superscope_reverse_buttons",
+                "snes9x_superscope_crosshair",
+                "snes9x_superscope_color",
+                "snes9x_justifier1_crosshair",
+                "snes9x_justifier1_color",
+                "snes9x_justifier2_crosshair",
+                "snes9x_justifier2_color",
+                "snes9x_rifle_crosshair",
+                "snes9x_rifle_color"
+            };
+
+            option_display.visible = show_lightgun_settings;
+
+            for (i = 0; i < 10; i++)
+            {
+                option_display.key = lightgun_keys[i];
+                environ_cb(RETRO_ENVIRONMENT_SET_CORE_OPTIONS_DISPLAY, &option_display);
+            }
+        }
+    }
+
+    var.key = "snes9x_show_advanced_av_settings";
+    var.value = NULL;
+
+    if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE, &var) && var.value)
+    {
+        bool show_advanced_av_settings_prev = show_advanced_av_settings;
+
+        show_advanced_av_settings = true;
+        if (strcmp(var.value, "disabled") == 0)
+            show_advanced_av_settings = false;
+
+        if (show_advanced_av_settings != show_advanced_av_settings_prev)
+        {
+            size_t i;
+            struct retro_core_option_display option_display;
+            char av_keys[15][32] = {
+                "snes9x_layer_1",
+                "snes9x_layer_2",
+                "snes9x_layer_3",
+                "snes9x_layer_4",
+                "snes9x_layer_5",
+                "snes9x_gfx_clip",
+                "snes9x_gfx_transp",
+                "snes9x_sndchan_1",
+                "snes9x_sndchan_2",
+                "snes9x_sndchan_3",
+                "snes9x_sndchan_4",
+                "snes9x_sndchan_5",
+                "snes9x_sndchan_6",
+                "snes9x_sndchan_7",
+                "snes9x_sndchan_8"
+            };
+
+            option_display.visible = show_advanced_av_settings;
+
+            for (i = 0; i < 15; i++)
+            {
+                option_display.key = av_keys[i];
+                environ_cb(RETRO_ENVIRONMENT_SET_CORE_OPTIONS_DISPLAY, &option_display);
+            }
+        }
+    }
 }
 
 static void S9xAudioCallback(void*)
@@ -648,6 +793,10 @@ float get_aspect_ratio(unsigned width, unsigned height)
     {
         return SNES_4_3;
     }
+    else if (aspect_ratio_mode == ASPECT_RATIO_4_3_SCALED)
+    {
+        return (4.0f * (MAX_SNES_HEIGHT - height)) / (3.0f * (MAX_SNES_WIDTH - width));
+    }
     else if (aspect_ratio_mode == ASPECT_RATIO_1_1)
     {
         return (float) width / (float) height;
@@ -687,10 +836,10 @@ void retro_get_system_av_info(struct retro_system_av_info *info)
 
     info->geometry.base_width = width;
     info->geometry.base_height = height;
-    info->geometry.max_width = MAX_SNES_WIDTH;
+    info->geometry.max_width = MAX_SNES_WIDTH_NTSC;
     info->geometry.max_height = MAX_SNES_HEIGHT;
     info->geometry.aspect_ratio = get_aspect_ratio(width, height);
-    info->timing.sample_rate = 32000;
+    info->timing.sample_rate = 32040;
     info->timing.fps = retro_get_region() == RETRO_REGION_NTSC ? 21477272.0 / 357366.0 : 21281370.0 / 425568.0;
 
     g_screen_gun_width = width;
@@ -990,19 +1139,17 @@ bool retro_load_game(const struct retro_game_info *game)
 
     if (rom_loaded)
     {
-        int pixel_format = RGB555;
-        if (environ_cb)
+        /* If we're in RGB565 format, switch frontend to that */
+        if (RED_SHIFT_BITS == 11)
         {
             enum retro_pixel_format fmt = RETRO_PIXEL_FORMAT_RGB565;
-            if (environ_cb(RETRO_ENVIRONMENT_SET_PIXEL_FORMAT, &fmt))
-                pixel_format = RGB565;
+            if (!environ_cb || !environ_cb(RETRO_ENVIRONMENT_SET_PIXEL_FORMAT, &fmt))
+            {
+                return false;
+            }
         }
 
-        S9xGraphicsDeinit();
-        S9xSetRenderPixelFormat(pixel_format);
-        S9xGraphicsInit();
-
-        update_geometry();
+        g_geometry_update = true;
 
         if (randomize_memory)
         {
@@ -1137,19 +1284,14 @@ bool retro_load_game_special(unsigned game_type, const struct retro_game_info *i
 
     if (rom_loaded)
     {
-        int pixel_format = RGB555;
-        if(environ_cb)
+        if(RED_SHIFT_BITS == 11)
         {
-            pixel_format = RGB565;
             enum retro_pixel_format fmt = RETRO_PIXEL_FORMAT_RGB565;
-            if (!environ_cb(RETRO_ENVIRONMENT_SET_PIXEL_FORMAT, &fmt))
-                pixel_format = RGB555;
+            if (!environ_cb || !environ_cb(RETRO_ENVIRONMENT_SET_PIXEL_FORMAT, &fmt))
+                return false;
         }
-        S9xGraphicsDeinit();
-        S9xSetRenderPixelFormat(pixel_format);
-        S9xGraphicsInit();
 
-        update_geometry();
+        g_geometry_update = true;
     }
 
     return rom_loaded;
@@ -1202,14 +1344,15 @@ void retro_init(void)
     Settings.FrameTimeNTSC = 16667;
     Settings.SixteenBitSound = TRUE;
     Settings.Stereo = TRUE;
-    Settings.SoundPlaybackRate = 32000;
-    Settings.SoundInputRate = 32000;
+    Settings.SoundPlaybackRate = 32040;
+    Settings.SoundInputRate = 32040;
     Settings.SupportHiRes = TRUE;
     Settings.Transparency = TRUE;
     Settings.AutoDisplayMessages = TRUE;
     Settings.InitialInfoStringTimeout = 120;
     Settings.HDMATimingHack = 100;
     Settings.BlockInvalidVRAMAccessMaster = TRUE;
+    Settings.SeparateEchoBuffer = FALSE;
     Settings.CartAName[0] = 0;
     Settings.CartBName[0] = 0;
     Settings.AutoSaveDelay = 1;
@@ -1232,9 +1375,11 @@ void retro_init(void)
     S9xSetSoundMute(FALSE);
     S9xSetSamplesAvailableCallback(S9xAudioCallback, NULL);
 
-    GFX.Pitch = MAX_SNES_WIDTH * sizeof(uint16);
+    GFX.Pitch = MAX_SNES_WIDTH_NTSC * sizeof(uint16);
     screen_buffer = (uint16*) calloc(1, GFX.Pitch * (MAX_SNES_HEIGHT + 16));
     GFX.Screen = screen_buffer + (GFX.Pitch >> 1) * 16;
+    ntsc_screen_buffer = (uint16*) calloc(1, GFX.Pitch * (MAX_SNES_HEIGHT + 16));
+    snes_ntsc_buffer = ntsc_screen_buffer + (GFX.Pitch >> 1) * 16;
     S9xGraphicsInit();
 
     S9xInitInputDevices();
@@ -1290,6 +1435,11 @@ static int scope_buttons[] =
 };
 static int scope_button_count = sizeof( scope_buttons ) / sizeof( int );
 
+#define SUPER_SCOPE_TRIGGER 2
+#define SUPER_SCOPE_CURSOR 3
+#define SUPER_SCOPE_TURBO 4
+#define SUPER_SCOPE_START 5
+
 #define JUSTIFIER_TRIGGER 2
 #define JUSTIFIER_START 3
 #define JUSTIFIER_OFFSCREEN 4
@@ -1378,8 +1528,8 @@ static void input_report_gun_position( unsigned port, int s9xinput )
 {
 	int x, y;
 
-	x = input_state_cb(port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_SCREEN_X);
-	y = input_state_cb(port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_SCREEN_Y);
+    x = input_state_cb(port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_SCREEN_X);
+    y = input_state_cb(port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_SCREEN_Y);
 
 	/*scale & clamp*/
 	x = ( ( x + 0x7FFF ) * g_screen_gun_width ) / 0xFFFF;
@@ -1398,6 +1548,144 @@ static void input_report_gun_position( unsigned port, int s9xinput )
 	S9xReportPointer(s9xinput, (int16_t)x, (int16_t)y);
 }
 
+static void input_handle_pointer_lightgun( unsigned port, unsigned gun_device, int s9xinput )
+{
+    int x, y;
+    x = input_state_cb(port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_X);
+    y = input_state_cb(port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_Y);
+
+	/*scale & clamp*/
+	x = ( ( x + 0x7FFF ) * g_screen_gun_width ) / 0xFFFF;
+	if ( x < 0 )
+		x = 0;
+	else if ( x >= g_screen_gun_width )
+		x = g_screen_gun_width - 1;
+
+	/*scale & clamp*/
+	y = ( ( y + 0x7FFF ) * g_screen_gun_height ) / 0xFFFF;
+	if ( y < 0 )
+		y = 0;
+	else if ( y >= g_screen_gun_height )
+		y = g_screen_gun_height - 1;
+
+    // Touch sensitivity: Keep the gun position held for a fixed number of cycles after touch is released
+    // because a very light touch can result in a misfire
+    if ( pointer_cycles_after_released > 0 && pointer_cycles_after_released < POINTER_PRESSED_CYCLES ) {
+        pointer_cycles_after_released++;
+        x = pointer_pressed_last_x;
+        y = pointer_pressed_last_y;
+        S9xReportPointer(s9xinput, (int16_t)x, (int16_t)y);
+        return;
+    }
+
+    if ( input_state_cb( port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_PRESSED ) )
+    {
+        pointer_pressed = 1;
+        pointer_cycles_after_released = 0;
+        pointer_pressed_last_x = x;
+        pointer_pressed_last_y = y;
+    } else if ( pointer_pressed ) {
+        pointer_cycles_after_released++;
+        pointer_pressed = 0;
+        x = pointer_pressed_last_x;
+        y = pointer_pressed_last_y;
+        // unpress the primary trigger
+        switch (gun_device)
+        {
+        case RETRO_DEVICE_LIGHTGUN_SUPER_SCOPE:
+            S9xReportButton(MAKE_BUTTON(PAD_2, setting_superscope_reverse_buttons ? SUPER_SCOPE_CURSOR : SUPER_SCOPE_TRIGGER), false);
+            break;
+        case RETRO_DEVICE_LIGHTGUN_JUSTIFIER:
+            S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_TRIGGER), false);
+            break;
+        case RETRO_DEVICE_LIGHTGUN_MACS_RIFLE:
+            S9xReportButton(MAKE_BUTTON(PAD_2, MACS_RIFLE_TRIGGER), false);
+            break;
+        default:
+            break;
+        }
+        return;
+    }
+    S9xReportPointer(s9xinput, (int16_t)x, (int16_t)y);
+
+    // triggers
+    switch (gun_device)
+    {
+        case RETRO_DEVICE_LIGHTGUN_SUPER_SCOPE:
+        {
+            bool start_pressed = false;
+            bool trigger_pressed = false;
+            bool turbo_pressed = false;
+            bool cursor_pressed = false;
+            if ( input_state_cb(port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_PRESSED) ) {
+                int touch_count = input_state_cb(port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_COUNT);
+                if ( touch_count == 4 ) {
+                    // start button
+                    start_pressed = true;
+                } else if ( touch_count == 3 ) {
+                    turbo_pressed = true;
+                } else if ( touch_count == 2 ) {
+                    if ( setting_superscope_reverse_buttons )
+                    {
+                        trigger_pressed = true;
+                    } else
+                    {
+                        cursor_pressed = true;
+                    }
+                } else {
+                    if ( setting_superscope_reverse_buttons )
+                    {
+                        cursor_pressed = true;
+                    } else
+                    {
+                        trigger_pressed = true;
+                    }
+                }
+            }
+            S9xReportButton(MAKE_BUTTON(PAD_2, SUPER_SCOPE_START), start_pressed);
+            S9xReportButton(MAKE_BUTTON(PAD_2, SUPER_SCOPE_TRIGGER), trigger_pressed);
+            S9xReportButton(MAKE_BUTTON(PAD_2, SUPER_SCOPE_CURSOR), cursor_pressed);
+            bool old_turbo = turbo_pressed;
+            turbo_pressed = turbo_pressed && !snes_superscope_turbo_latch;
+            snes_superscope_turbo_latch = old_turbo;
+            S9xReportButton(MAKE_BUTTON(PAD_2, SUPER_SCOPE_TURBO), turbo_pressed);
+            break;
+        }
+
+        case RETRO_DEVICE_LIGHTGUN_JUSTIFIER:
+        {
+            bool trigger_pressed = false;
+            bool start_pressed = false;
+            bool offscreen = false;
+            if ( input_state_cb(port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_PRESSED) ) {
+                int touch_count = input_state_cb(port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_COUNT);
+                if ( touch_count == 3 ) {
+                    start_pressed = true;
+                } else if ( touch_count == 2 ) {
+                    offscreen = true;
+                } else {
+                    trigger_pressed = true;
+                }
+            }
+            S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_TRIGGER), trigger_pressed || offscreen);
+            S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_START), start_pressed ? 1 : 0 );
+            S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_OFFSCREEN), offscreen);
+            break;
+        }
+        case RETRO_DEVICE_LIGHTGUN_MACS_RIFLE:
+        {
+            int pressed = input_state_cb(port, RETRO_DEVICE_POINTER, 0, RETRO_DEVICE_ID_POINTER_PRESSED);
+            S9xReportButton(MAKE_BUTTON(PAD_2, MACS_RIFLE_TRIGGER),pressed);
+            break;
+        }
+        case RETRO_DEVICE_NONE:
+            break;
+        default:
+            if (log_cb)
+                log_cb(RETRO_LOG_ERROR, "Unknown device for touchscreen lightgun...\n");
+    }
+}
+
 static void report_buttons()
 {
     int offset = snes_devices[0] == RETRO_DEVICE_JOYPAD_MULTITAP ? 4 : 1;
@@ -1430,81 +1718,101 @@ static void report_buttons()
 
             case RETRO_DEVICE_LIGHTGUN_SUPER_SCOPE:
 
-				input_report_gun_position( port, BTN_POINTER );
+                if ( setting_gun_input == SETTING_GUN_INPUT_POINTER ) {
+                    input_handle_pointer_lightgun(port, RETRO_DEVICE_LIGHTGUN_SUPER_SCOPE, BTN_POINTER);
+                } else {
+                    // Lightgun is default
+                    input_report_gun_position( port, BTN_POINTER );
 
-				for (int i = 0; i < scope_button_count; i++)
-				{
-					int id = scope_buttons[i];
-					bool btn = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, id )?true:false;
+                    for (int i = 0; i < scope_button_count; i++)
+                    {
+                        int id = scope_buttons[i];
+                        bool btn = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, id )?true:false;
 
-					/* RETRO_DEVICE_ID_LIGHTGUN_TURBO special case - core needs a rising-edge trigger */
-					if ( id == RETRO_DEVICE_ID_LIGHTGUN_TURBO )
-					{
-						bool old = btn;
-						btn = btn && !snes_superscope_turbo_latch;
-						snes_superscope_turbo_latch = old;
-					}
-
-					S9xReportButton(MAKE_BUTTON(PAD_2, i+2), btn);
-				}
+                        /* RETRO_DEVICE_ID_LIGHTGUN_TURBO special case - core needs a rising-edge trigger */
+                        if ( id == RETRO_DEVICE_ID_LIGHTGUN_TURBO )
+                        {
+                            bool old = btn;
+                            btn = btn && !snes_superscope_turbo_latch;
+                            snes_superscope_turbo_latch = old;
+                        }
+                        int super_scope_button_id = i+2;
+                        if ( setting_superscope_reverse_buttons )
+                        {
+                            if ( super_scope_button_id == SUPER_SCOPE_TRIGGER ) {
+                                super_scope_button_id = SUPER_SCOPE_CURSOR;
+                            } else if ( super_scope_button_id == SUPER_SCOPE_CURSOR ) {
+                                super_scope_button_id = SUPER_SCOPE_TRIGGER;
+                            }
+                        }
+                        S9xReportButton(MAKE_BUTTON(PAD_2, super_scope_button_id), btn);
+                    }
+                }
                 break;
 
             case RETRO_DEVICE_LIGHTGUN_JUSTIFIER:
 
-				input_report_gun_position( port, BTN_POINTER );
+                if ( setting_gun_input == SETTING_GUN_INPUT_POINTER ) {
+                    input_handle_pointer_lightgun(port, RETRO_DEVICE_LIGHTGUN_JUSTIFIER, BTN_POINTER);
+                } else {
+                    // Lightgun is default
+                    input_report_gun_position( port, BTN_POINTER );
 
-				{
-					/* Special Reload Button */
-					int btn_offscreen_shot = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_RELOAD );
+                    {
+                        /* Special Reload Button */
+                        int btn_offscreen_shot = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_RELOAD );
 
-					/* Trigger ? */
-					int btn_trigger = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_TRIGGER );
-					S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_TRIGGER), btn_trigger || btn_offscreen_shot);
+                        /* Trigger ? */
+                        int btn_trigger = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_TRIGGER );
+                        S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_TRIGGER), btn_trigger || btn_offscreen_shot);
 
-					/* Start Button ? */
-					int btn_start = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_START );
-					S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_START), btn_start ? 1 : 0 );
+                        /* Start Button ? */
+                        int btn_start = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_START );
+                        S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_START), btn_start ? 1 : 0 );
 
-					/* Aiming off-screen ? */
-					int btn_offscreen = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_IS_OFFSCREEN );
-					S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_OFFSCREEN), btn_offscreen || btn_offscreen_shot);
-				}
+                        /* Aiming off-screen ? */
+                        int btn_offscreen = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_IS_OFFSCREEN );
+                        S9xReportButton(MAKE_BUTTON(PAD_2, JUSTIFIER_OFFSCREEN), btn_offscreen || btn_offscreen_shot);
+                    }
 
-				/* Second Gun? */
-				if ( snes_devices[port+1] == RETRO_DEVICE_LIGHTGUN_JUSTIFIER_2 )
-				{
-					int second = port+1;
+                    /* Second Gun? */
+                    if ( snes_devices[port+1] == RETRO_DEVICE_LIGHTGUN_JUSTIFIER_2 )
+                    {
+                        int second = port+1;
 
-					input_report_gun_position( second, BTN_POINTER2 );
+                        input_report_gun_position( second, BTN_POINTER2 );
 
-					/* Special Reload Button */
-					int btn_offscreen_shot = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_RELOAD );
+                        /* Special Reload Button */
+                        int btn_offscreen_shot = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_RELOAD );
 
-					/* Trigger ? */
-					int btn_trigger = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_TRIGGER );
-					S9xReportButton(MAKE_BUTTON(PAD_3, JUSTIFIER_TRIGGER), btn_trigger || btn_offscreen_shot);
+                        /* Trigger ? */
+                        int btn_trigger = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_TRIGGER );
+                        S9xReportButton(MAKE_BUTTON(PAD_3, JUSTIFIER_TRIGGER), btn_trigger || btn_offscreen_shot);
 
-					/* Start Button ? */
-					int btn_start = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_START );
-					S9xReportButton(MAKE_BUTTON(PAD_3, JUSTIFIER_START), btn_start ? 1 : 0 );
-
-					/* Aiming off-screen ? */
-					int btn_offscreen = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_IS_OFFSCREEN );
-					S9xReportButton(MAKE_BUTTON(PAD_3, JUSTIFIER_OFFSCREEN), btn_offscreen || btn_offscreen_shot);
-				}
+                        /* Start Button ? */
+                        int btn_start = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_START );
+                        S9xReportButton(MAKE_BUTTON(PAD_3, JUSTIFIER_START), btn_start ? 1 : 0 );
 
+                        /* Aiming off-screen ? */
+                        int btn_offscreen = input_state_cb( second, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_IS_OFFSCREEN );
+                        S9xReportButton(MAKE_BUTTON(PAD_3, JUSTIFIER_OFFSCREEN), btn_offscreen || btn_offscreen_shot);
+                    }
+                }
                 break;
 
             case RETRO_DEVICE_LIGHTGUN_MACS_RIFLE:
 
- 				input_report_gun_position( port, BTN_POINTER );
-
-				{
-					/* Trigger ? */
-					int btn_trigger = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_TRIGGER );
-					S9xReportButton(MAKE_BUTTON(PAD_2, MACS_RIFLE_TRIGGER), btn_trigger);
-				}
+                if ( setting_gun_input == SETTING_GUN_INPUT_POINTER ) {
+                    input_handle_pointer_lightgun(port, RETRO_DEVICE_LIGHTGUN_MACS_RIFLE, BTN_POINTER);
+                } else {
+                    input_report_gun_position( port, BTN_POINTER );
 
+                    {
+                        /* Trigger ? */
+                        int btn_trigger = input_state_cb( port, RETRO_DEVICE_LIGHTGUN, 0, RETRO_DEVICE_ID_LIGHTGUN_TRIGGER );
+                        S9xReportButton(MAKE_BUTTON(PAD_2, MACS_RIFLE_TRIGGER), btn_trigger);
+                    }
+                }
                 break;
 
             case RETRO_DEVICE_NONE:
@@ -1523,7 +1831,8 @@ void retro_run()
     bool updated = false;
     if (environ_cb(RETRO_ENVIRONMENT_GET_VARIABLE_UPDATE, &updated) && updated)
         update_variables();
-    if (height != PPU.ScreenHeight)
+
+    if (g_geometry_update || height != PPU.ScreenHeight)
     {
         update_geometry();
         height = PPU.ScreenHeight;
@@ -1558,6 +1867,7 @@ void retro_deinit()
     S9xUnmapAllControls();
 
     free(screen_buffer);
+    free(ntsc_screen_buffer);
 }
 
 
@@ -1670,6 +1980,7 @@ bool retro_unserialize(const void* data, size_t size)
 
 bool8 S9xDeinitUpdate(int width, int height)
 {
+    static int burst_phase = 0;
     int overscan_offset = 0;
 
     if (crop_overscan_mode == OVERSCAN_CROP_ON)
@@ -1708,7 +2019,18 @@ bool8 S9xDeinitUpdate(int width, int height)
     }
 
 
-    if (width == MAX_SNES_WIDTH && hires_blend)
+    if (blargg_filter)
+    {
+        burst_phase = (burst_phase + 1) % 3;
+
+        if (width == 512)
+            snes_ntsc_blit_hires(snes_ntsc, GFX.Screen, GFX.Pitch / 2, burst_phase, width, height, snes_ntsc_buffer, GFX.Pitch);
+        else
+            snes_ntsc_blit(snes_ntsc, GFX.Screen, GFX.Pitch / 2, burst_phase, width, height, snes_ntsc_buffer, GFX.Pitch);
+
+        video_cb(snes_ntsc_buffer + ((int)(GFX.Pitch >> 1) * overscan_offset), SNES_NTSC_OUT_WIDTH(256), height, GFX.Pitch);
+    }
+    else if (width == MAX_SNES_WIDTH && hires_blend)
     {
         #define AVERAGE_565(el0, el1) (((el0) & (el1)) + ((((el0) ^ (el1)) & 0xF7DE) >> 1))
 
@@ -1801,18 +2123,15 @@ const char* S9xGetDirectory(s9x_getdirtype type)
 {
     switch (type)
     {
-        case ROMFILENAME_DIR:
-            return g_rom_dir;
         case BIOS_DIR:
             return retro_system_directory;
         default:
-            break;
+            return g_rom_dir;
     }
 
     return "";
 }
 void S9xInitInputDevices() {}
-const char* S9xChooseFilename(unsigned char) { return ""; }
 void S9xHandlePortCommand(s9xcommand_t, short, short) {}
 bool S9xPollButton(unsigned int, bool*) { return false; }
 void S9xToggleSoundChannel(int) {}
@@ -1822,12 +2141,10 @@ bool8 S9xInitUpdate() { return TRUE; }
 void S9xExtraUsage() {}
 bool8 S9xOpenSoundDevice() { return TRUE; }
 bool S9xPollAxis(unsigned int, short*) { return FALSE; }
-void S9xSetPalette() {}
 void S9xParseArg(char**, int&, int) {}
 void S9xExit() {}
 bool S9xPollPointer(unsigned int, short*, short*) { return false; }
 
-const char *S9xChooseMovieFilename(unsigned char) { return NULL; }
 void S9xMessage(int type, int, const char* s)
 {
     if (!log_cb) return;
@@ -1880,66 +2197,3 @@ void S9xAutoSaveSRAM()
 {
     return;
 }
-
-#ifndef __WIN32__
-// S9x weirdness.
-void _splitpath (const char *path, char *drive, char *dir, char *fname, char *ext)
-{
-    *drive = 0;
-
-    const char	*slash = strrchr(path, SLASH_CHAR),
-            *dot	= strrchr(path, '.');
-
-    if (dot && slash && dot < slash)
-        dot = NULL;
-
-    if (!slash)
-    {
-        *dir = 0;
-
-        strcpy(fname, path);
-
-        if (dot)
-        {
-            fname[dot - path] = 0;
-            strcpy(ext, dot + 1);
-        }
-        else
-            *ext = 0;
-    }
-    else
-    {
-        strcpy(dir, path);
-        dir[slash - path] = 0;
-
-        strcpy(fname, slash + 1);
-
-        if (dot)
-        {
-            fname[dot - slash - 1] = 0;
-            strcpy(ext, dot + 1);
-        }
-        else
-            *ext = 0;
-    }
-}
-
-void _makepath (char *path, const char *, const char *dir, const char *fname, const char *ext)
-{
-    if (dir && *dir)
-    {
-        strcpy(path, dir);
-        strcat(path, SLASH_STR);
-    }
-    else
-        *path = 0;
-
-    strcat(path, fname);
-
-    if (ext && *ext)
-    {
-        strcat(path, ".");
-        strcat(path, ext);
-    }
-}
-#endif // __WIN32__
