@@ -680,6 +680,20 @@ void S9xNPSendServerPause (bool8 paused)
     S9xNPSendToAllClients (pause, 7);
 }
 
+void S9xNPSendJoypadSwap()
+{
+#ifdef NP_DEBUG
+	printf("SERVER: Swap Joypads - @%ld\n", S9xGetMilliTime() - START);
+#endif
+	uint8 swap[7];
+	uint8 *ptr = swap;
+	*ptr++ = NP_SERV_MAGIC;
+	*ptr++ = 0;
+	*ptr++ = NP_SERV_JOYPAD_SWAP;
+	WRITE_LONG(ptr, 7);
+	S9xNPSendToAllClients(swap, 7);
+}
+
 void S9xNPServerLoop (void *)
 {
 #ifdef __WIN32__
@@ -1155,8 +1169,11 @@ void S9xNPSendSRAMToClient (int c)
     uint8 sram [7];
     int SRAMSize = Memory.SRAMSize ?
                    (1 << (Memory.SRAMSize + 3)) * 128 : 0;
-    if (SRAMSize > 0x10000)
-        SRAMSize = 0x10000;
+    if (Memory.LoROM)
+        SRAMSize = SRAMSize < 0x70000 ? SRAMSize : 0x70000;
+	else if (Memory.HiROM)
+		SRAMSize = SRAMSize < 0x40000 ? SRAMSize : 0x40000;
+
     int len = 7 + SRAMSize;
 
     sprintf (NetPlay.ActionMsg, "SERVER: Sending S-RAM to player %d...", c + 1);
